# -*- coding: utf-8 -*-

import numpy as np
from matplotlib import pyplot as plt
from warnings import warn
from tdyno.cnst import Cnst
from tdyno.s2t import S2T
from tdyno.pct import PCT
from tdyno.sc_tp_t import Hm, HP, Gsn
from tdyno.sc2t import PS2, TSS2
from tdyno.sc2t_md import TSSMD2
from tdyno.rt2 import RT2
from tdyno.mnt.mnt_t_fa import MntMltPntAmp
from tdyno.mnt.mnt_t_poyn import Mnt2DSqPoynU
from tdyno.mnt.mnt_t_md import MntWgMdU


# todo: API use frequency instead of omega. (however, k contains 2pi anyway?)

class TDyno:
    def __init__(self):
        """
        Finite-difference time-domain (FDTD) package with dynamic modulations in the refractive index and/or in gain/loss.

        """
        self.c = Cnst()
        self.st: S2T = None
        self.scs = []
        self.pc = None
        self.mnts = []
        self.rt = None  # RT2
        self.polarization = 'Ez'
        self.dt = None
        self.Nt = None
        self.tps = []
        self.if_ndc = False
        self.omg_ndc = None
        self.md = 'HDE'

    def setup(self, xmin, xmax, dx, ymin, ymax, dy, epsi, mu, x=None, y=None, polarization='Ez', dt=None, Nt=5000, mode='BHDE', if_ndc=False, ndc_omega=1.):
        """
        Set up the solving space.

        The space is assumed to be in the x-y plane. z is out-of-plane.

        Parameters
        ----------
        xmin, xmax, dx      :   float
                                x limits and grid size
        ymin, ymax, dy      :   float
                                y limits and grid size
        epsi, mu            :   float
                                background permittivity and permeability
        x, y                :   ndarray
                                1d ndarray, If supplied, override xmin etc.
        dt                  :   float
                                time step in simulation. If simulation diverges, change to a smaller number.
        polarization        :   str
                                ['Ez'|'Hz'], 'Ez' means Ez, Hx, Hy mode. 'Hz' means Hz, Ex, Ey mode.
        Nt                  :   int
                                Total run time. After it stops, you can click "Play/Pause" again to run another Nt time steps.
        mode                :   str
                                ['BHDE'|'HDE']
        if_ndc              :   bool
                                if using numeric dispersion correction
        ndc_omega           :   float
                                frequency for ndc

        Returns
        -------

        """

        if not dt:
            dt = 1. / 2. * np.sqrt(dx ** 2 + dy ** 2)

        self.st: S2T = S2T(xmin, xmax, dx, ymin, ymax, dy, epsi, mu, x=x, y=y)
        self.set_polarization(polarization)
        self.set_dt(dt)
        self.set_Nt(Nt)
        self.solving_mode(mode)
        self.use_ndc(if_ndc, ndc_omega)

    def add_structure(self, shape='rectangle',
                      xmin=None, xmax=None, ymin=None, ymax=None,
                      xcenter=None, ycenter=None, radius=None, angle1=None, angle2=None, width=None,
                      kind='dielectric',
                      epsi=None, mu=None, conductivity=0.,
                      res_omegas=None, d_epsi=None, gammas=None,
                      m_amp=None, m_omega=None, m_q=None, m_phase=0.):

        """
        Add structures to the solving space.

        `shape` can be one of the following:
        'rectangle'|'circular'|'ring'|'wedge'
        For 'rectangle',    specify `xmin`, `xmax`, `ymin`, `ymax`.
        For 'disk',         specify centers `xcenter`, `ycenter`, and `radius`.
        For 'ring',         specify `xcenter`, `ycenter`, outer `radius`, and `width`.
        For 'wedge',        specify `xcenter`, `ycenter`, outer `radius`, `width`, and `angle1` and `angle2`.

        The type of material for the structure is specified in `kind`, which can be one of the following:

        'dielectric'|'Lorentz'|'metal'|'index modulated'|'gain-loss modulated'|‘Lorentz modulated'

        'dielectric' means dispersionless dielectric. It needs the following material parameters:
            `epsi`, `mu`, `conductivity` (optional).

        'Lorentz' means dielectric with Lorentz model. It needs the following material parameters:
            `epsi`, `mu`, `res_omegas`, `d_epsi`, `gammas`.

        'metal' means metallic material with Drude model. It needs the following material parameters:
            `epsi`, `mu`, `res_omegas`, `gammas`.

        'index modulated' means dynamic modulation in the refractive index. It needs the following material parameters:
            `epsi`, `mu`, `m_amp`, `m_omega`, `m_q`, `m_phase`.

        'loss-gain modulated' means dynamic modulation in the loss and gain. It needs the following material parameters:
            `epsi`, `mu`, 'conductivity' (optional), `m_amp`, `m_omega`, `m_q`, `m_phase`.

        'Lorentz modulated' means Lorentz model with modulated resonance strengh. It needs the following material parameters:
            `epsi`, `mu`, `res_omegas`, `d_epsi`, `gammas`, `m_amp`, `m_omega`, `m_q`, `m_phase`.

        Parameters
        ----------
        shape                   :   str
                                    ['rectangle'|'disk'|'ring'|'wedge'】
        xmin, xmax, ymin, ymax  :   float
                                    four corners of the structure. If shape is rectangular, these four parameters should be supplied.
        xcenter, ycenter        :   float
                                    x and y center of structure, for use in circular and ring
        radius                  :   float
        angle1, angle2          :   float
                                    start and end angle of wedge, in degrees, between 0 to 360
        width                   :   float
                                    width of ring or wedge
        kind                    :   str
                                    ['dielectric'|'Lorentz'|'metal'|'index modulated'|'loss-gain modulated'|‘Lorentz modulated']
        epsi, mu                :   float or Tuple[float, float, float] or 3x3 ndarray
                                    The permittivity and permeability.
                                    For Lorentz dielectric and Metal material, these two are the values at the high-frequency limit.
        res_omegas              :   list[float]
                                    resonance frequencies
        d_epsi                  :   list[float]
                                    permittivity change at each resonance, i.e. oscillator strength
        Gammas                  :   list[float]
                                    Broadening of each resonance
        conductivity            :   float
                                    loss (positive number) or gain (negative) in the material
        m_amp                   :   float or list[float]
                                    modulation amplitude
        m_omega                 :   float or list[float]
                                    modulation frequency
        m_q                     :   Tuple[float, float] or list[Tuple[float, float]]
                                    (qx, qy), wave vector of modulation.
        m_phase                 :   float or list[float]
                                    modulation phase. By default, cos modulation.

        Returns
        -------

        """
        if shape == 'rectangle':
            shape = 'rct'
        elif shape == 'disk':
            shape = 'ccl'
        elif shape == 'ring':
            shape = 'rng'
        elif shape == 'wedge':
            shape = 'wdg'
        else:
            warn('The shape of the added structure is not recognized', UserWarning)

        if kind == 'dielectric':
            if not conductivity:
                kind = 'sp'
            else:
                kind = 'lg'
        elif kind == 'Lorentz':
            kind = 'Lz'
        elif kind == 'metal':
            kind = 'Dr'
        elif kind == 'index modulated':
            kind = 'dmri'
        elif kind == 'gain-loss modulated':
            kind = 'dmlg'
        elif kind == 'Lorentz modulated':
            kind = 'dmLz'
        else:
            warn('Material type not recognized.', UserWarning)

        self.st.a_b(shp=shape,
                    xmin_b=xmin, xmax_b=xmax, ymin_b=ymin, ymax_b=ymax,
                    xc_b=xcenter, yc_b=ycenter,
                    r_b=radius,
                    a_1=angle1, a_2=angle2,
                    w_b=width,
                    knd=kind,
                    epsi_b=epsi, mu_b=mu,
                    epsi_infty_b=epsi, mu_infty_b=mu,
                    omgs_rsn=res_omegas, dlts_epsi=d_epsi, Gms=gammas,
                    sgm=conductivity,
                    m_a=m_amp, m_o=m_omega, m_q=m_q, m_p=m_phase)

    def set_polarization(self, polarization):
        """
        Choose polarization of simulation.

        Parameters
        ----------
        polarization        :   str
                                ['Ez'|'Hz']
                                'Ez' means Ez, Hx, Hy mode. 'Hz' means Hz, Ex, Ey mode.

        Returns
        -------

        """

        if polarization == 'Ez' or 'Hz':
            self.polarization = polarization
        else:
            warn("Polarization must be either 'Ez' or 'Hz'.", UserWarning)

    def set_dt(self, dt):
        """

        Parameters
        ----------
        dt      :   float
                    time step in simulation. If simulation diverges, change to a smaller number.

        Returns
        -------

        """

        self.dt = dt

    def set_Nt(self, Nt):
        """
        set total run time

        Parameters
        ----------
        Nt      :   int
                    Total run time. After it stops, you can click "Play/Pause" again to run another Nt time steps.
        Returns
        -------

        """

        self.Nt = Nt

    def add_source_temporal(self, kind, omega=1., tau=1., phase=0., t_before=None, t_span=0.):
        """
        set up the temporal profile for sources.

        For any source, its temporal profile can be any of the following
        `cw` (continuous wave),
            Need to specify `omega`, `phase` (optional), `tau` (optional),  `t_before` (optional)
            There will be a ramping phase in the beginning of cw wave to avoid simulation diverging.
            `tau` is the ramping time constant.
            `t_before` is the total time before reaching constant magnitude in cw.
        `pulse` (single Gaussian pulse),
            Need to specify `tau`, `t_before` (optional)
            `tau is the width of the pulse.
            `t_before` is the total time before the peak of the pulse in the simulation.
        `packet` (cw wave modulated by a Gaussian pulse)
            Need to specify `omega`, `phase` (optional), `tau`, `t_before`, `t_span`.
            `tau` is the width of the pulse profile.
            `t_before` is the total time before reaching the peak in amplitude.
            `t_span` is the amount of time the wave stay at its peak amplitude, i.e. the wave packet can be such that there is a cw section between the ramping-up and the ramping-down stages.

        Parameters
        ----------
        kind            :   str
                            ['pulse'|'cw'|'packet']
        omega           :   float
                            frequency
        tau             :   float
                            time-constant for Gaussian pulse, or the wave packet
        phase           :   float
                            phase shift of the wave
        t_before        :   float
                            total time before the peak of the Gaussian. default to be 6 times of tau
        t_span          :   float
                            when the amplitude reaches the peak of a wave packet, maintain t_span time of cw

        Returns
        -------
        t_profile       :   Gsn or Hm or HP
                            the temporal profile object

        """
        t_profile = None
        if kind == 'pulse':
            t_profile = Gsn(self.c, tau, t_bf=t_before)
        elif kind == 'cw':
            t_profile = Hm(self.c, omega, phi=phase, t_bf=t_before, t_r=tau)
        elif kind == 'packet':
            t_profile = HP(self.c, omega, phi=phase, t_spn=t_span, t_bf=t_before, t_r=tau)
        else:
            warn('Temporal profile not recognized.', UserWarning)

        self.tps.append(t_profile)
        return t_profile

    def use_ndc(self, if_use_ndc, omega):
        """
        Use numeric dispersion correction in the main solver. This corrects the phase velocity of waves in vacuum to be light speed.

        Notes
        -----
        This is not to be confused with the numeric dispersion correction in sources. If numeric dispersion correction is used in sources, then the user shouldn't enable this in the main solver.

        In general, increasing spatial resolution reduces numeric dispersion.

        Parameters
        ----------
        if_use_ndc          :   bool
        omega               :   float
                                frequency for ndc

        Returns
        -------

        """
        if if_use_ndc:
            self.if_ndc = True
            self.omg_ndc = omega
            if self.rt:
                self.rt.set_ndc(if_use_ndc, omega)
        else:
            self.if_ndc = False

    def add_point_source(self, x, y, amp, t_profile):
        """
        Add a point source in the space

        Parameters
        ----------
        x, y        :   float
                        position of the point source
        amp         :   float
                        amplitude of the source
        t_profile   :   Hm or HP or Gsn
                        source temporal profile

        -------

        """
        sc = PS2(x, y, amp, t_profile, self.st, plrz=self.polarization)

        self.scs.append(sc)

    def add_tfsf_source(self, xmin, xmax, ymin, ymax,
                        kx, ky, amp, t_profile,
                        epsi=1., mu=1.,
                        if_ndc=False, omega_ndc=None,
                        sides='all'):
        """
        Add a TFSF source in the space

        Parameters
        ----------
        xmin, xmax ymin ymax    :   float
                                    four corners of TFSF source region
        kx, ky                  :   float
                                    kx and ky determine incident direction.

                                    The magnitudes of kx and ky don't matter, only the direction matters.

                                    The incident wave is assumed to be travelling in free space so it's k is determined by its frequency.

        amp                     :   float
                                    amplitude of the incident wave
        t_profile               :   Hm or Gsn or HP
                                    source temporal profile
        epsi, mu                :   float
        if_ndc                  :   bool
                                    if use numeric dispersion correction. If enabled, the phase velocity of the source fields will be the exact numerical one, not the analytic light speed in the space.
        omega_ndc               :   float
                                    the phase velocity of the source field will be the numeric phase velocity at this frequency
        sides                   :   str
                                    {'all', combinations of 't', 'b', 'l', 'r'}

                                    Controls which sides of the TFSF source exist.

                                    Default to be 'all', i.e. TFSF source in a rectangular region.

                                    Can choose any combination of 't', 'b', 'l', 'r' for top, bottom, left and right.

                                    If choose any ONE of the sides, it becomes a uni-directional plane-wave source.

        Returns
        -------

        """

        sc = TSS2(xmin, xmax, ymin, ymax, kx, ky, amp, t_profile, self.st, self.c.c0, self.dt, plrz=self.polarization, if_ndc=if_ndc, omg_ndc=omega_ndc, epsi=epsi, mu=mu, whr=sides)

        self.scs.append(sc)

    def add_tfsf_mode_source(self, xmin, xmax, ymin, ymax,
                             kx, ky, amp, t_profile,
                             xi, f, epsi, mu, omega,
                             xi0=None, reverse_direction=False,
                             if_ndc=False,
                             sides='all'):
        """
        Add a TFSF source of a waveguide mode.

        Parameters
        ----------
        xmin, xmax, ymin, ymax  :   float
                                    four corners of TFSF source region
        kx, ky                  :   float
                                    (kx, ky) determine the wave propagation constant. Both the magnitude and the direction of (kx, ky) matters. This is different from `add_tfsf_source`.
        amp                     :   float
                                    amplitude of the incident wave
        t_profile               :   Hm or Gsn or HP
                                    source temporal profile
        xi                      :   array_like
                                    1d array, spatial points where f is defined. See notes on `f` below.
        f                       :   array_like
                                    1d array value of f, the waveguide mode.

                                    For Ez mode, `f` is Ez. For Hz mode, `f` is Hz.

                                    `xi` and `f` can be analytically calculated or numerically simulated.

                                    `xi` are the actual physical points where `f` is defined.
                                    For Ez mode, `xi` is where Ez is defined. For Hz mode, `xi` is where Hz is defined.
                                    In fact `xi` does not know about Yee cells in general (for example, analytically simulated).

                                    If `xi` and `f` were simulation results of `tdyno` with a waveguide in the x direction, then for `Ez` polarization `xi` are the Yee cell corners i.e. grid points in y, while for `Hz` polarization `xi` are the half grid points in y in each Yee cell.

        epsi                    :   array_like
                                    2d array. The permittivity profile along the transverse direction of the waveguide.

                                    `epsi[i]` defines the values at index i. It has 2 elements, `epsi[i][0]` is the component  along the waveguide direction, `epsi[i][1]` is transverse to the waveguide.

        mu                      :   array_like
                                    2d array. The permeability profile along the transverse direction of the waveguide.

                                    `mu[i]` defines the values at index i. It has 2 elements, `mu[i][0]` is the component along the waveguide direction, `mu[i][1]` is transverse to the waveguide.

        omega                   :   float
                                    intended frequency of the waveguide mode. This frequency will also be used for ndc, i.e. the phase velocity of the source field will be the numeric phase velocity at this frequency
        xi0                     :   float
                                    signed distance from reference line (xi=0) to origin
        reverse_direction       :   bool
                                    If `True`, reverse `beta` direction, while keeping profile `f` unchanged.

                                    Can instead manually set kx and ky to negative, in which case it is rotating both `beta` and `f` 180 degrees. Note, for asymmetric waveguide, modal profile `f` is asymmetric.
        if_ndc                  :   bool
                                    if use numeric dispersion correction. If enabled, the phase velocity of the source fields will be the exact numerical one, not the analytic light speed in the space.

                                    If True, do not use ndc in the main update equations.

        sides                   :   str
                                    {'all', combinations of 't', 'b', 'l', 'r'}

                                    Controls which sides of the TFSF source exist.

                                    Default to be 'all', i.e. TFSF source in a rectangular region.

                                    Can choose any combination of 't', 'b', 'l', 'r' for top, bottom, left and right.

                                    If choose any ONE of the sides, it becomes a uni-directional plane-wave source.

        Notes
        -----
        The source does not know the actual structure in the solving space. This is to say, if the structure is the intended waveguide then this mode will be generated in the TF/SF fashion.

        To generate the waveguide mode, the permittivity and permeability are needed and are supplied. Again, this is regardless of the actual structure in the solving space.

        """

        sc = TSSMD2(xmin, xmax, ymin, ymax, kx, ky, amp, t_profile, xi, f, epsi, mu, omega, self.st, self.c.c0, self.dt, xi0=xi0, reverse_direction=reverse_direction, plrz=self.polarization, if_ndc=if_ndc, omg_ndc=omega, whr=sides)

        self.scs.append(sc)

    def add_pml(self, poly_ord, ratio, amax_factor=1., kappa=1., npx=0, npy=0):
        """
        Add PML to the space

        Parameters
        ----------
        poly_ord        :   float or Tuple[float, float, float]
                            polynomial order. Tuple of 3 if want different polynomial order for x, y, and z direction.
        ratio           :   float or Tuple[float, float, float]
                            target attenuation ratio
        amax_factor     :   float
                            a_max factor. Adjusted to reduce low frequency reflectance. Coefficient a_max is computed automatically, but one can manually scale this factor.
        kappa           :   float or Tuple[float, float, float]
                            kappa, the kappa parameter stretches the real part of xyz in PML.
                            If the material itself has some absorption, this helps with attenuation.
                            However, this is at the cost of all partial derivatives in the update equations will have to multiplied with the kappa matrix (it is some value in PML but is 1 in the solving domain).
        npx, npy        :   int
                            number of pml cells in x and y (npx=10 means 10 pml cells on each end).


        Returns
        -------

        """

        self.pc = PCT(self.dt, poly_ord, ratio, a_max_fct=amax_factor, kpp=kappa, Nx=self.st.Nx, Ny=self.st.Ny, dx=self.st.dx, dy=self.st.dy, npx=npx, npy=npy, npz=0, epsi0=self.c.epsi0, mu0=self.c.mu0)

    def solving_mode(self, mode):
        """
        Choose the solving mode.
        If permeability is 1 throughout the space, choose 'HDE' for faster solving.

        Parameters
        ----------
        mode        :   str
                        'HDE' or 'BHDE'

        Returns
        -------

        """
        self.md = mode

    def add_point_monitors(self, coords, delay=0, weights=None, omega_min=0., omega_max=2., n_omega=100, norm_factor=1., flux=False, refer_spectrum=None):
        """
        Add monitors to monitor the amplitude of the field on multi points. Record the amplitudes on each, do weighted sum, and do Fourier Transform.
        The monitors only record Ez and Hz for the two corresponding polarizations, respectively.

        Parameters
        ----------
        delay                   :   int
                                    time delay. Won't start Fourier Transform until this time step.
        coords                  :   list[Tuple[float, float]]
                                    coordinates of the points to monitor. Each element is a tuple of (x, y)
        weights                 :   list[float]
                                    the amplitudes of the monitored points are multiplied by these weights, and then add up
        omega_min, omega_max    :   float
        n_omega                 :   int
                                    number of frequency points
        norm_factor             :   float
                                    normalization factor for spectrum
        flux                    :   bool
                                    if true, plot the flux spectral density. Otherwise plot the power spectral density
        refer_spectrum          :   ndarray[float]
                                    reference spectrum. If supplied, the shown spectrum will be divided by this. This is for convenience, for example, if want to see transmission spectrum.

        Returns
        -------

        """
        if flux is True:
            show = 'flux spectral density'
        else:
            show = 'energy spectral density'
        mnt = MntMltPntAmp(self.st, coords, dt=self.dt, td=delay, wts=weights, omin=omega_min, omax=omega_max, n_o=n_omega, nmf=norm_factor, show=show, ref_spctrm=refer_spectrum)
        self.mnts.append(mnt)

    def add_poynting_monitor(self,
                             xmin=None, xmax=None, ymin=None, ymax=None,
                             delay=None,
                             omega_min=None, omega_max=None, n_omega=None,
                             norm_factor=None, refer_spectrum=None,
                             sides='all'):
        """
        Add Poynting flux monitor as a square box. Tracks the total energy flux from inside the box to outside in real time and calculates the spectrum.

        Parameters
        ----------
        xmin : float
        xmax : float
        ymin : float
        ymax : float
        delay : int
            delay starting monitor by this many time steps
        omega_min : float
        omega_max : float
        n_omega : int
        norm_factor : float
        refer_spectrum : np.ndarray
        sides : str
            Can be "all", or any combinations (any order) of 'l', 'r', 't', 'b'.

        Returns
        -------

        """
        mnt = Mnt2DSqPoynU(st=self.st,
                           xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
                           dt=self.dt, td=delay,
                           omin=omega_min, omax=omega_max, n_o=n_omega,
                           nmf=norm_factor, ref_spctrm=refer_spectrum,
                           whr=sides, plrz=self.polarization)
        self.mnts.append(mnt)

    def add_mode_monitor(self,
                         x=None, y=None,
                         xmin=None, xmax=None, ymin=None, ymax=None,
                         ):
        """
        A monitor to record modal profile at an interface in either x or y.

        Parameters
        ----------
        x : float
            x position of the interface if it's in y
        y : float
            y position of the interface if it's in x
        xmin, xmax, ymin, ymax : float
            min and max of the interface. Default to the edge of the solving space.

        Notes
        -----
        If `x` is supplied, `ymin` and `ymax` are needed. If `y is supplied, `xmin` and `xmax` are needed. If both `x` and `y` are supplied, `y` is ignored.
        """

        mnt = MntWgMdU(st=self.st,
                       x=x, y=y,
                       xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
                       plrz=self.polarization)
        self.mnts.append(mnt)

    def run(self, skipping=5, vmin=-1., vmax=1., **kwargs):
        """
        Run the FDTD simulation.

        Parameters
        ----------
        skipping        :   int
                            only plot the fields once every `skipping` time steps, to save time
        vmin, vmax      :   float
                            the colormap range for the field strength in plotting
        kwargs          ：  dict
                            key word arguments for PlotField2DinAx

        Returns
        -------

        """

        self.rt = RT2(self.c.c0, self.c.epsi0, self.c.mu0,
                      self.st,
                      self.scs,
                      self.dt, self.Nt,
                      pc=self.pc,
                      plrz=self.polarization,
                      if_ndc=self.if_ndc, omg_ndc=self.omg_ndc,
                      md=self.md,
                      mnts=self.mnts,
                      vmin=vmin, vmax=vmax,
                      skp=skipping,
                      **kwargs)

        plt.show()



# -*- coding: utf-8 -*-

import numpy as np
from warnings import warn
# import time

from tdyno.s2t import S2T
from tdyno.mnt.mnt_t_md_ui import MntMdUI


class MntWgMdU:
    def __init__(self, *args, **kwargs):
        """
        wrapper class, `MntWgMd` with `MntMdUI`
        """
        self.mnt: MntWgMd = MntWgMd(*args, **kwargs)
        self.mui: MntMdUI = MntMdUI(xi=self.mnt.xi, epsi=self.mnt.epsi, mu=self.mnt.mu, f=self.mnt.f, plrz=self.mnt.plrz)

    @property
    def pchs(self):
        return self.mnt.pchs

    def rnf(self, *args, **kwargs):
        self.mnt.rnf(*args, **kwargs)

    def up(self):
        self.mui.up(f=self.mnt.f)

    def cf(self):
        self.mnt.cf()


class MntWgMd:

    def __init__(self,
                 st,
                 x=None, y=None,
                 xmin=None, xmax=None, ymin=None, ymax=None,
                 plrz=None):
        """
        Monitor waveguide mode.

        At any time, record the instantaneous Ez or Hz values, the coordinate transverse to the waveguide, permittivity and permeability values at these coordinates.

        Can feed new field data into the monitor through `rnf()`. If polarization is "Ez", the arguments to `rnf()` is understood as Ez, Hx, Hy. For "Hz" polarization, it is understood as Hz, Ex, Ey.

        Only works for waveguides with simple dielectric materials.

        Parameters
        ----------
        st: S2T
        x : float
        y : float
        xmin : float
        xmax : float
        ymin : float
        ymax : float
        plrz : str

        Notes
        -----
        If `x` is specified, then need `ymin` and `ymax`. It set up an interface in y direction at `x`.

        If `y` is specified, then need `xmin` and `xmax`. It set up an interface in x direction at `y`.

        If both `x` and `y` are given, `y` is ignored.
        """

        self._st = st
        if (plrz != 'Ez') and (plrz != 'Hz'):
            warn("polarization not recognized, default to 'Hz' (TE).", UserWarning)
            plrz = 'Hz'
        self.plrz = plrz

        xmin, xmax, ymin, ymax = [smm if mm is None else mm
                                  for (mm, smm) in zip([xmin, xmax, ymin, ymax],
                                                       [st.xmin, st.xmax, st.ymin, st.ymax])]

        xx_n, yy_n = [st.xx_n, st.yy_n]
        self._xx_n, self._yy_n = (xx_n, yy_n)
        xmin_n, xmax_n, ymin_n, ymax_n = [int(np.floor(cor / dcor)) for cor, dcor in
                                          zip([xmin + st.dx / 1e4 - st.xmin, xmax + st.dx / 1e4 - st.xmin, ymin + st.dx / 1e4 - st.ymin, ymax + st.dx / 1e4 - st.ymin],
                                              [st.dx, st.dx, st.dy, st.dy])]

        if xmin_n < xx_n.min():
            warn('Source left boundary outside of the solving space. It has been reset to the left edge of the solving space.')
            xmin_n = xx_n.min()
        if xmax_n > xx_n.max():
            warn('Source right boundary outside of the solving space. It has been reset to the right edge of the solving space.')
            xmax_n = xx_n.max()
        if ymin_n < yy_n.min():
            warn('Source bottom boundary outside of the solving space. It has been reset to the bottom edge of the solving space.')
            ymin_n = yy_n.min()
        if ymax_n > yy_n.max():
            warn('Source top boundary outside of the solving space. It has been reset to the top edge of the solving space.')
            ymax_n = yy_n.max()
        self._xmin_n, self._xmax_n, self._ymin_n, self._ymax_n = [xmin_n, xmax_n, ymin_n, ymax_n]

        # indexing of edge in the bulk
        if x is not None:
            x_n = int(np.floor(((x + 1e-10 - st.xmin) / st.dx)))
            self._i_e = (xx_n == x_n) * (yy_n < ymax_n) * (yy_n >= ymin_n)

            xmi = x
            ymi = ymin
            width = 0.
            height = ymax - ymin

            if y is not None:
                warn("Both `x` and `y` are specified. y is ignored, x is retained.", UserWarning)
        else:
            if y is None:
                warn("Neither `x` or `y` is specified. Assume `y = 0`.")
                y = 0.

            y_n = int(np.floor(((y + 1e-10 - st.ymin) / st.dy)))
            self._i_e = (yy_n == y_n) * (xx_n < xmax_n) * (xx_n >= xmin_n)

            xmi = xmin
            ymi = y
            width = xmax - xmin
            height = 0.

        pch = {'shp': 'rct',
               'xy': (xmi, ymi),
               'width': width,
               'height': height
               }
        self.pchs = []
        self.pchs.append(pch)

        self.xi = self._st.yy[self._i_e]
        if self.plrz == "Hz":
            self.xi += self._st.dy / 2.

        embg = [em[[0, 1], [0, 1]] for em in [self._st.epsi_bg, self._st.mu_bg]]
        em = [a[None, None, :] * np.ones((self._st.Ny, self._st.Nx))[:, :, None] for a in embg]
        for bxsp in self._st.bxs_sp:
            for a, b in zip(em, [bxsp.mtr.epsi, bxsp.mtr.mu]):
                a[bxsp.gmt.iib, :] = b[[0, 1], [0, 1]]
        self.epsi, self.mu = [a[self._i_e, :] for a in em]

        self._fz = None
        self.f = None

    def rnf(self, fz, fx, fy):
        """
        Receive new fields

        Parameters
        ----------
        fx : np.ndarray
        fy : np.ndarray
        fz : np.ndarray
            all of these inputs have shape (Ny, Nx)
        """

        self._fz = fz
        # self._fx = fx
        # self._fy = fy

        self.f = fz[self._i_e]

    def cf(self):
        self._fz = None
        self.f = None



# -*- coding: utf-8 -*-


import numpy as np
from warnings import warn
from scipy import sparse as sps


class PCT:

    def __init__(self, dt, p, rto, a_max_fct=1., kpp=1., Nx=1, Ny=1, Nz=1, dx=1., dy=1., dz=1., npx=0, npy=0, npz=0, epsi0=1., mu0=1.):

        """
        PML coefficients for FDTD.

        Parameters
        ----------
        dt              :   float
        p               :   float or Tuple[float, float, float]
                            polynomial order. List of 3 if want different polynomial order for x, y, and z direction.
        rto             :   float or Tuple[float, float, float]
                            target attenuation ratio
        a_max_fct       :   float
                            a_max factor. Coefficient a_max is computed based on sigma(dx), but one can manually scale this factor.
        kpp             :   float or Tuple[float, float, float]
                            kappa, the kappa parameter stretches the real part of xyz in PML.
        Nx, Ny, Nz      :   int
        dx, dy, dz      :   float
                            The solving space parameters.
        npx, npy, npz   :   int
                            number of pml cells in x, y and z (npx=10 means 10 pml cells on each end).
        epsi0, mu0      :   float
        """

        if type(p) is float:
            self.px, self.py, self.pz = [p, p, p]
        else:
            self.px, self.py, self.pz = p

        if type(rto) is float:
            self.rto_x, self.rto_y, self.rto_z = [rto, rto, rto]
        else:
            self.rto_x, self.rto_y, self.rto_z = rto

        if type(kpp) is float:
            kpp = [kpp, kpp, kpp]
        kpp_x, kpp_y, kpp_z = kpp
        kpp_inv = [1. / ka for ka in kpp]

        self.kix = np.ones([Ny, Nx, Nz], dtype=float)
        self.kiy = np.ones([Ny, Nx, Nz], dtype=float)
        self.kiz = np.ones([Ny, Nx, Nz], dtype=float)
        self.kix[:, :npx, :] = kpp_inv[0]
        self.kix[:, (Nx - npx):, :] = kpp_inv[0]
        self.kiy[:npy, :, :] = kpp_inv[1]
        self.kiy[(Ny - npy):, :, :] = kpp_inv[1]
        self.kiz[:, :, :npz] = kpp_inv[2]
        self.kiz[:, :, (Nz - npz):] = kpp_inv[2]
        self.kixV, self.kiyV, self.kizV = [s.reshape(Nx * Ny * Nz) for s in [self.kix, self.kiy, self.kiz]]

        # coefficients sgm, a, b and c
        sgm_xb_a, a_xb_a, b_xb_a, c_xb_a = self.c_pc_1_a('b', dt, p, rto, a_max_fct, kpp_x, dx, Nx, npx, epsi0, mu0)
        sgm_xf_a, a_xf_a, b_xf_a, c_xf_a = self.c_pc_1_a('f', dt, p, rto, a_max_fct, kpp_x, dx, Nx, npx, epsi0, mu0)
        sgm_yb_a, a_yb_a, b_yb_a, c_yb_a = self.c_pc_1_a('b', dt, p, rto, a_max_fct, kpp_y, dy, Ny, npy, epsi0, mu0)
        sgm_yf_a, a_yf_a, b_yf_a, c_yf_a = self.c_pc_1_a('f', dt, p, rto, a_max_fct, kpp_y, dy, Ny, npy, epsi0, mu0)
        sgm_zb_a, a_zb_a, b_zb_a, c_zb_a = self.c_pc_1_a('b', dt, p, rto, a_max_fct, kpp_z, dz, Nz, npz, epsi0, mu0)
        sgm_zf_a, a_zf_a, b_zf_a, c_zf_a = self.c_pc_1_a('f', dt, p, rto, a_max_fct, kpp_z, dz, Nz, npz, epsi0, mu0)

        # coefficients b and c, in x y and z direction
        self.bxb, self.byb, self.bzb = np.meshgrid(b_xb_a, b_yb_a, b_zb_a)
        self.bxf, self.byf, self.bzf = np.meshgrid(b_xf_a, b_yf_a, b_zf_a)
        self.cxb, self.cyb, self.czb = np.meshgrid(c_xb_a, c_yb_a, c_zb_a)
        self.cxf, self.cyf, self.czf = np.meshgrid(c_xf_a, c_yf_a, c_zf_a)

        # self.bxbM, self.bybM, self.bzbM, self.bxfM, self.byfM, self.bzfM, \
        # self.cxbM, self.cybM, self.czbM, self.cxfM, self.cyfM, self.czfM \
        #     = [sps.diags(s.reshape(Nx*Ny*Nz), 0, shape=(Nx*Ny*Nz, Nx*Ny*Nz))
        #        for s in [self.bxb, self.byb, self.bzb, self.bxf, self.byf, self.bzf,
        #                  self.cxb, self.cyb, self.czb, self.cxf, self.cyf, self.czf]]

        # flatten the coefficient matrices.
        self.bxbV, self.bybV, self.bzbV, self.bxfV, self.byfV, self.bzfV, \
        self.cxbV, self.cybV, self.czbV, self.cxfV, self.cyfV, self.czfV \
            = [s.reshape(Nx*Ny*Nz) for s in [self.bxb, self.byb, self.bzb, self.bxf, self.byf, self.bzf,
                                             self.cxb, self.cyb, self.czb, self.cxf, self.cyf, self.czf]]

        x_n = np.arange(Nx)
        y_n = np.arange(Ny)
        z_n = np.arange(Nz)
        xx_n, yy_n, zz_n = np.meshgrid(x_n, y_n, z_n)

        # indexing of PML regions
        ixiP = np.logical_or(xx_n < npx, xx_n >= (Nx - npx))
        iyiP = np.logical_or(yy_n < npy, yy_n >= (Ny - npy))
        iziP = np.logical_or(zz_n < npz, zz_n >= (Nz - npz))
        self.iiP = np.logical_or(ixiP, iyiP, iziP)
        self.iiPV = self.iiP.ravel()
        # self.iiPM = np.diagflat(self.iiP)

        # b and c coefficients
        self.bxbiP, self.bybiP, self.bzbiP, self.bxfiP, self.byfiP, self.bzfiP, \
        self.cxbiP, self.cybiP, self.czbiP, self.cxfiP, self.cyfiP, self.czfiP \
            = [coef[self.iiP] for coef in [self.bxb, self.byb, self.bzb, self.bxf, self.byf, self.bzf,
                                           self.cxb, self.cyb, self.czb, self.cxf, self.cyf, self.czf]]
        self.bxbViP, self.bybViP, self.bzbViP, self.bxfViP, self.byfViP, self.bzfViP, \
        self.cxbViP, self.cybViP, self.czbViP, self.cxfViP, self.cyfViP, self.czfViP \
            = [coef[self.iiPV] for coef in [self.bxbV, self.bybV, self.bzbV, self.bxfV, self.byfV, self.bzfV,
                                            self.cxbV, self.cybV, self.czbV, self.cxfV, self.cyfV, self.czfV]]

    def c_pc_1_a(self, fb, dt, p, rto, a_max_fct, kpp, dL, N, Np, epsi0, mu0):

        """
        Calculate PML coefficients as 1d array.

        Parameters
        ----------
        fb              :   str
                            'f' or 'b'
        dt              :   float
        p               :   int
        rto             :   float
                            ratio
        a_max_fct       :   float
                            a_max factor
        kpp             :   float
                            kappa
        dL              :   float
        N               :   int
        Np              :   int
                            number of PML cells
        epsi0
        mu0

        Returns
        -------
        sgm         :   float
        a           :   ndarray
        b           :   ndarray
        c           :   ndarray
        """

        a = np.zeros(N, dtype=float)
        sgm = np.zeros(N, dtype=float)
        c = np.zeros(N, dtype=float)

        if Np != 0:
            dp = Np * dL
            # vacuum impedance
            vac_imp = np.sqrt(mu0 / epsi0)
            sgm_max = - (p + 1) * np.log(rto) / 2. / vac_imp / dp
            a_max = a_max_fct * 1000 * sgm_max * (1. / 2 / N) ** p

            if fb == 'f':
                # sigma as a function of l
                sgml = sgm_max * ((np.arange(Np) + 0.5) / float(Np)) ** p
                sgm[(N - Np):] = sgml
                sgm[(Np - 1)::-1] = sgml

                # a as a function l
                al = a_max * (1 - (np.arange(Np) + 0.5) / float(Np)) ** p
                a[(N - Np):] = al
                a[(Np - 1)::-1] = al

            elif fb == 'b':
                sgml_r = sgm_max * (np.arange(Np) / float(Np)) ** p
                sgml_l = sgm_max * ((np.arange(Np) + 1.) / float(Np)) ** p
                sgm[(N - Np):] = sgml_r
                sgm[(Np - 1)::-1] = sgml_l

                al_r = a_max * (1 - np.arange(Np) / float(Np)) ** p
                al_l = a_max * (1 - (np.arange(Np) + 1.) / float(Np)) ** p
                a[(N - Np):] = al_r
                a[(Np - 1)::-1] = al_l
            else:
                warn('fb must be f or b!', UserWarning)

            b = np.exp(-1. / epsi0 * (sgm / kpp + a) * dt)

            # indexing of pml region.
            ip = (sgm != 0)
            c[ip] = sgm[ip] / (sgm[ip] * kpp + kpp ** 2 * a[ip]) * (b[ip] - 1)

        else:
            b = np.zeros(N, dtype=float)
            c = np.zeros(N, dtype=float)

        return sgm, a, b, c



import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.widgets import Button, TextBox
from matplotlib.gridspec import GridSpec
from warnings import warn


class MntUI1S1S:
    def __init__(self,
                 signal_x_label=None,
                 signal_y_label=None,
                 spectrum_x_label=None,
                 spectrum_y_label=None,
                 sx=None, sy=None,
                 Sx=None, Sy=None
                 ):
        """
        Monitor UI, one signal, one spectrum.

        Parameters
        ----------
        signal_y_label : str
        spectrum_y_label : str
        sx : array_like
        sy : array_like
            sx and sy are the x and y axis of the "signal"
        Sx : array_like
        Sy : array_like
            Sx and Sy are the x and y axis of the "spectrum"
        """

        self.sx = sx
        self.sy = sy
        self.Sx = Sx
        self.Sy = Sy

        # Start plotter
        mpl.rcParams['mathtext.fontset'] = 'cm'
        fontsize_label = 20
        fontsize_axis = 16
        gs = GridSpec(nrows=2, ncols=2, height_ratios=[1, 15])

        if signal_x_label is None:
            signal_x_label = 'Time'
        if signal_y_label is None:
            signal_y_label = 'Signal'
        if spectrum_x_label is None:
            spectrum_x_label = "$\omega$"
        if spectrum_y_label is None:
            spectrum_y_label = "Energy spectral density"
        self.fig = plt.figure(figsize=(14, 5))
        ax_f = self.fig.add_subplot(gs[1, 0])
        ax_f.set_xlabel(signal_x_label, fontsize=fontsize_label)
        ax_f.set_ylabel(signal_y_label, fontsize=fontsize_label)
        ax_f.tick_params(axis='both', which='major', labelsize=fontsize_axis)
        self.ax_f = ax_f
        x, y = [s if (s is not None) else [] for s in [sx, sy]]
        self.ln_f = ax_f.plot(x, y)[0]

        ax_F = self.fig.add_subplot(gs[1, 1])
        ax_F.set_xlabel(spectrum_x_label, fontsize=fontsize_label)
        ax_F.set_ylabel(spectrum_y_label, fontsize=fontsize_label)
        ax_F.tick_params(axis='both', which='major', labelsize=fontsize_axis)
        self.ax_F = ax_F
        x, y = [s if (s is not None) else [] for s in [Sx, Sy]]
        if Sx is None:
            x = []
        else:
            x = Sx
            if Sy is None:
                y = np.zeros(len(Sx))
        self.ln_F = ax_F.plot(x, y)[0]
        if hasattr(Sx, '__len__'):
            ax_F.set_xlim([Sx[0], Sx[-1]])

        gs.tight_layout(self.fig)

        self.rec_n = 0
        self.spct_n = 0

        # save recording button
        self.ax_sv_rec = plt.axes([0.05, 0.93, 0.1, 0.05])
        self.b_sv_rec = Button(self.ax_sv_rec, r'Save recording')
        self.b_sv_rec.on_clicked(self.sv_rec)

        # save recording title
        self.rec_ttl = 'rec'
        self.ax_rec_ttl = plt.axes([0.21, 0.93, 0.27, 0.05])
        self.tb_rec_ttl = TextBox(self.ax_rec_ttl, 'file name', initial=self.rec_ttl)
        self.tb_rec_ttl.on_text_change(self.set_rec_ttl)

        # save spectrum button
        self.ax_sv_spct = plt.axes([0.55, 0.93, 0.1, 0.05])
        self.b_sv_spct = Button(self.ax_sv_spct, r'Save spectrum')
        self.b_sv_spct.on_clicked(self.sv_spct)

        # save recording title
        self.spct_ttl = 'spectrum'
        self.ax_spct_ttl = plt.axes([0.71, 0.93, 0.27, 0.05])
        self.tb_spct_ttl = TextBox(self.ax_spct_ttl, 'file name', initial=self.spct_ttl)
        self.tb_spct_ttl.on_text_change(self.set_spct_ttl)

    def up(self,
           sx=None, sy=None,
           Sx=None, Sy=None):
        """
        Update plots

        Parameters
        ----------
        sx : array_like
        sy : array_like
        Sx : array_like
        Sy : array_like

        Returns
        -------

        """
        self.sx, self.sy, self.Sx, self.Sy = [s if (s is not None) else ss for s, ss in zip([sx, sy, Sx, Sy], [self.sx, self.sy, self.Sx, self.Sy])]

        if (sx is not None) or (sy is not None):
            self.ln_f.set_data(self.sx, self.sy)
            self.ax_f.set_xlim([self.sx[0], self.sx[-1]])
            self.ax_f.set_ylim([min(self.sy), max(self.sy)])
            self.ax_f.draw_artist(self.ax_f.patch)
            self.ax_f.draw_artist(self.ln_f)

        if Sx is not None:  # most likely this does not change
            self.ln_F.set_xdata(Sx)
            self.ax_F.set_xlim([Sx[0], Sx[-1]])
            self.ax_F.draw_artist(self.ax_F.patch)
            self.ax_F.draw_artist(self.ln_F)
        if Sy is not None:
            self.ln_F.set_ydata(Sy)
            self.ax_F.set_ylim([min(Sy), max(Sy)])
            self.ax_F.draw_artist(self.ax_F.patch)
            self.ax_F.draw_artist(self.ln_F)

        self.fig.canvas.flush_events()
        self.fig.canvas.update()

    def sv_rec(self, event):
        """
        save recorded monitor values

        Returns
        -------

        """

        data = np.concatenate([[self.sx], [self.sy]], axis=0).T
        ttl = self.rec_ttl + '_{:03d}.txt'.format(self.rec_n)
        np.savetxt(ttl, data)
        self.rec_n += 1

    def set_rec_ttl(self, tx):
        """
        set file title for recorded monitor values

        Parameters
        ----------
        tx      :   str

        Returns
        -------

        """
        self.rec_ttl = tx
        self.rec_n = 0

    def sv_spct(self, event):
        """
        save monitor spectrum

        Returns
        -------

        """

        data = np.concatenate([[self.Sx], [self.Sy]], axis=0).T
        ttl = self.spct_ttl + '_{:03d}.txt'.format(self.spct_n)
        np.savetxt(ttl, data)
        self.spct_n += 1

    def set_spct_ttl(self, tx):
        """
        set file title for monitor spectrum

        Parameters
        ----------
        tx      :   str

        Returns
        -------

        """
        self.spct_ttl = tx
        self.spct_n = 0



# -*- coding: utf-8 -*-

import numpy as np
import matplotlib as mpl
import matplotlib.axes as axes
import matplotlib.pyplot as plt
from matplotlib.widgets import Button, TextBox
from matplotlib.gridspec import GridSpec
from warnings import warn


class MntMdUI:
    def __init__(self,
                 xi, epsi, mu,
                 x_label=None,
                 y_label=None,
                 f=None,
                 plrz=None
                 ):
        """
        Monitor UI for waveguide mode.

        Parameters
        ----------
        x_label, y_label : str
        xi : array_like
            1d array
        f : array_like
            1d array
        epsi, mu : array_like
            2d array of shape (len(xi), 2)
        plrz : str
            only needed for file name. if not supplied, it is ignored.
        """

        if f is None:
            f = np.zeros(len(xi))

        self.xi = xi
        self.f = f
        self.epsi = epsi
        self.mu = mu

        if plrz is None:
            plrz = ''
        self.plrz = plrz

        self.mm = 0

        # Start plotter
        mpl.rcParams['mathtext.fontset'] = 'cm'
        fontsize_label = 20
        fontsize_axis = 16
        gs = GridSpec(nrows=2, ncols=2, height_ratios=[1, 15])

        if x_label is None:
            x_label = r'$\xi$'
        if y_label is None:
            y_label = 'amplitude'

        _xi, _f, _epsi, _mu = [s if (s is not None) else [] for s in [xi, f, epsi, mu]]

        self.fig = plt.figure(figsize=(14, 5))
        ax_f = self.fig.add_subplot(gs[1, 0])
        ax_f.set_xlabel(x_label, fontsize=fontsize_label)
        ax_f.set_ylabel(y_label, fontsize=fontsize_label)
        ax_f.tick_params(axis='both', which='major', labelsize=fontsize_axis)
        self.ax_f = ax_f
        self.ln_f = ax_f.plot(_xi, _f)[0]
        if hasattr(_xi, '__len__'):
            ax_f.set_xlim([_xi[0], _xi[-1]])

        ax_em: axes.Axes = self.fig.add_subplot(gs[1, 1])
        ax_em.set_xlabel(x_label, fontsize=fontsize_label)
        ax_em.set_ylabel(r"$\varepsilon$, $\mu$", fontsize=fontsize_label)
        ax_em.tick_params(axis='both', which='major', labelsize=fontsize_axis)
        self.ax_em = ax_em
        self.lns_epmu = ax_em.plot(_xi, epsi, _xi, mu)
        ax_em.legend([r"$\varepsilon_x$", r"$\varepsilon_y$", r"$\mu_x$", r"$\mu_y$"], fontsize=fontsize_label)
        if hasattr(_xi, '__len__'):
            ax_em.set_xlim(left=_xi[0], right=_xi[-1])

        gs.tight_layout(self.fig)

        self.rec_n = 0

        # save button
        self.ax_sv_rec = plt.axes([0.05, 0.93, 0.1, 0.05])
        self.b_sv_rec = Button(self.ax_sv_rec, r'Save')
        self.b_sv_rec.on_clicked(self.sv_rec)

        # save title
        self.rec_ttl = 'mode'
        self.ax_rec_ttl = plt.axes([0.21, 0.93, 0.27, 0.05])
        self.tb_rec_ttl = TextBox(self.ax_rec_ttl, 'file name', initial=self.rec_ttl)
        self.tb_rec_ttl.on_text_change(self.set_rec_ttl)

    def up(self,
           f=None,
           ):
        """
        Update plots

        Parameters
        ----------
        f : array_like
        """

        if f is not None:
            self.f = f
            self.ln_f.set_ydata(self.f)
            mm = max(max(np.abs(self.f)), self.mm)
            self.mm = mm
            self.ax_f.set_ylim([-mm, mm])
            self.ax_f.draw_artist(self.ax_f.patch)
            self.ax_f.draw_artist(self.ln_f)

        self.fig.canvas.flush_events()
        self.fig.canvas.update()

    def sv_rec(self, event):
        """
        save recorded monitor values

        Returns
        -------

        """

        ttl = [self.rec_ttl + '_' + self.plrz + '_' + c + '_{:03d}.txt'.format(self.rec_n) for c in ["xi", "f", "epsi", "mu"]]

        for t, d in zip(ttl, [self.xi, self.f, self.epsi, self.mu]):
            np.savetxt(t, d)

        print("Data saved. See files: ")
        for t in ttl:
            print(t)

        self.rec_n += 1

    def set_rec_ttl(self, tx):
        """
        set file title for recorded monitor values

        Parameters
        ----------
        tx      :   str

        Returns
        -------

        """
        self.rec_ttl = tx
        self.rec_n = 0



# -*- coding: utf-8 -*-


import numpy as np


class SigSpct:
    def __init__(self, dt, td=None, omin=None, omax=None, n_o=None, nmf=None, ref_spctrm=None, n=None):
        """"
        Given a set of independent time-series signals, calculate their respective Fourier spectrum (FT) as the signals comes in.

        Signal value is fed in at each time step. The FT of the signal is updated as they come in. The FT, the energy (esd) and power spectral density (psd), the number flux spectrum (flx) and flux rate spectrum (flxr) are calculated.

        Parameters
        ----------
        dt : float
            time step length
        td : int
            time steps delay before starting to calculate the spectrum
        omin : float
            lower bound of omega
        omax : float
            upper bound of omega
        n_o : int
            number of frequency points
        nmf : float or array_like
            normalization factor for spectrum. If it is array, it means we have `n` signals in parallel, the length must be equal to `n`.
        ref_spctrm : np.ndarray
            reference energy spectral density spectrum to be divided by. If it's 1d array, it means we are monitoring one signal. if it is (n, n_o) shape, it means we are monitoring several signals in parallel.
        n : int
            number of signals in parallel
        """
        if dt is None:
            dt = 1
        self.dt = dt
        if td is None:
            td = 0.
        self.td = td
        self.nmf = nmf
        self.ref_spctrm = ref_spctrm
        self.omin = omin
        if omin is None:
            omin = np.pi / self.dt / 1000
        self.omin = omin
        if omax is None:
            omax = np.pi / self.dt
        self.omax = omax
        if n_o is None:
            n_o = 201
        self.n_o = n_o
        self.omg = np.linspace(self.omin, self.omax, n_o)
        if nmf is None:
            nmf = 1.
        self.nmf = nmf
        if ref_spctrm is None:
            ref_spctrm = 1.
        self.ref_spctrm = ref_spctrm
        if n is None:
            n = 1
        self.n = n

        self.t_n = []  # list of integer time steps, last element is current
        self.ts = []  # list of real time points, last element is current
        self._nt = 0   # next time step

        self.f = []  # real time signal

        self.knl = np.exp(-1j * self.omg * self.dt)  # kernel

        # Initialize the Fourier transform of the field
        self.reset()
        self.F = np.zeros((n, n_o)) * 1j  # FT
        self.esd = np.zeros((n, n_o))  # energy spectral density
        self.psd = np.zeros((n, n_o))  # power spectral density
        self.flx = np.zeros((n, n_o))  # integral number flux spectral density
        self.flxr = np.zeros((n, n_o))  # number flux rate spectral density

    def rnf(self, f):
        """
        Receive new signal

        Parameters
        ----------
        f : float or array_like
            if f is array, it must be 1d array, which means it is several signals in parallel.

        """
        if not hasattr(f, '__len__'):
            f = np.array([f])
        f = np.array(f)
        if f.ndim > 1:
            raise Exception('The signal fed to the spectrum analyzer can only be 1d array.')
        self.f.append(f)
        t = self._nt * self.dt

        if self._nt > self.td:
            self.F += (self.knl ** self._nt)[None, :] * f[:, None] * self.dt
            self.esd = np.square(np.abs(self.F)) / self.nmf / self.ref_spctrm
            self.psd = self.esd / (t - self.dt * self.td)
            self.flx = self.esd / self.omg
            self.flxr = self.flx / (t - self.dt * self.td)

        self.t_n.append(self._nt)
        self.ts.append(t)
        self._nt += 1

    def reset(self):
        """
        reset/initialize signals and spectra

        """
        self.ts = []
        self.t_n = []
        self._nt = 0
        self.f = []

        n_o = self.n_o
        n = self.n
        self.F = np.zeros((n, n_o)) * 1j
        self.esd = np.zeros((n, n_o))
        self.psd = np.zeros((n, n_o))
        self.flx = np.zeros((n, n_o))
        self.flxr = np.zeros((n, n_o))



# -*- coding: utf-8 -*-


import scipy.sparse as sps


class DOM:

    def __init__(self, Nx=1, Ny=1, Nz=1, dx=1., dy=1., dz=1.):
        """
        Differential Operator Matrices.

        Parameters
        ----------
        Nx, Ny, Nz  :   integer
                        number of unit-cell sites in x, y, and z direction
        dx, dy, dz  :   float
                        unit-cell sizes

        """

        eye_Nx = sps.eye(Nx)
        eye_Ny = sps.eye(Ny)
        eye_Nz = sps.eye(Nz)

        if Nx > 1:
            pxfr = sps.diags([-1, 1, 1], [0, 1, -Nx+1], shape=(Nx, Nx))
            pxbr = sps.diags([1, -1, -1], [0, -1, Nx-1], shape=(Nx, Nx))
        else:
            pxfr = sps.diags([0.], 0, shape=(1, 1))
            pxbr = sps.diags([0.], 0, shape=(1, 1))
        self.pxf = sps.kron(eye_Nz, sps.kron(eye_Ny, pxfr)) / dx
        self.pxb = sps.kron(eye_Nz, sps.kron(eye_Ny, pxbr)) / dx

        if Ny > 1:
            self.pyf = sps.kron(eye_Nz,  (sps.kron(sps.diags([1, -1, 1], [-Ny + 1, 0, 1], shape=(Ny, Ny)), eye_Nx))  ) / dy
            self.pyb = sps.kron(eye_Nz,  (sps.kron(sps.diags([-1, 1, -1], [-1, 0, Ny - 1], shape=(Ny, Ny)), eye_Nx))  ) / dy
        else:
            self.pyf = sps.kron(eye_Nz,  (sps.kron(sps.diags([0.], 0, shape=(1, 1)), eye_Nx))  ) / dy
            self.pyb = sps.kron(eye_Nz,  (sps.kron(sps.diags([0.], 0, shape=(1, 1)), eye_Nx))  ) / dy

        eye_NyNx = sps.kron(eye_Ny, eye_Nx)
        if Nz > 1:
            self.pzf = sps.kron(sps.diags([1, -1, 1], [-Nz+1, 0, 1], shape=(Nz, Nz)), eye_NyNx) / dz
            self.pzb = sps.kron(sps.diags([-1, 1, -1], [-1, 0, Nz-1], shape=(Nz, Nz)), eye_NyNx) / dz
        else:
            self.pzf = sps.kron(sps.diags([0.], 0, shape=(1, 1)), eye_NyNx) / dz
            self.pzb = sps.kron(sps.diags([0.], 0, shape=(1, 1)), eye_NyNx) / dz


class DOMH2D:

    def __init__(self, Nx=1, Ny=1, Nz=1, dx=1., dy=1., dz=1.):
        """
        Differential Operator Matrices Hexagonal 2D.

        Parameters
        ----------
        Nx, Ny, Nz      :   int
        dx, dy, dz      :   float
        """

        # self.Nx = Nx
        # self.Ny = Ny
        # self.P = Nx * Ny

        eye_Nx = sps.eye(Nx)
        eye_Ny = sps.eye(Ny)

        # Nx and Ny should be > 1
        if Nx > 1:
            pxfr = sps.diags([-1, 1, 1], [0, 1, -Nx + 1], shape=(Nx, Nx))
            pxbr = sps.diags([1, -1, -1], [0, -1, Nx - 1], shape=(Nx, Nx))
            self.pxf = sps.kron(eye_Ny, pxfr) / dx
            self.pxb = sps.kron(eye_Ny, pxbr) / dx
        if Ny > 1:
            self.pyf = (-sps.kron(eye_Ny, eye_Nx)
                        + sps.kron(sps.diags([1], [1], shape=(Ny, Ny)), eye_Nx)
                        + sps.kron(sps.diags([1], [-Ny + 1], shape=(Ny, Ny)), sps.diags([1, 1], [Nx / 2, -Nx / 2], shape=(Nx, Nx)))) / dy
            self.pyb = (sps.kron(eye_Ny, eye_Nx)
                        + sps.kron(sps.diags([-1], [-1], shape=(Ny, Ny)), eye_Nx)
                        + sps.kron(sps.diags([1], [Ny - 1], shape=(Ny, Ny)), sps.diags([-1, -1], [Nx / 2, -Nx / 2], shape=(Nx, Nx)))) / dy

        eye_NyNx = sps.kron(eye_Ny, eye_Nx)
        if Nz > 1:
            self.pzf = sps.kron(sps.diags([1, -1, 1], [-Nz+1, 0, 1], shape=(Nz, Nz)), eye_NyNx) / dz
            self.pzb = sps.kron(sps.diags([-1, 1, -1], [-1, 0, Nz-1], shape=(Nz, Nz)), eye_NyNx) / dz
        else:
            self.pzf = sps.kron(sps.diags([0.], 0, shape=(1, 1)), eye_NyNx) / dz
            self.pzb = sps.kron(sps.diags([0.], 0, shape=(1, 1)), eye_NyNx) / dz



# -*- coding: utf-8 -*-

import numpy as np
from warnings import warn
import time

from tdyno.s2t import S2T
from tdyno.mnt.sig_spct import SigSpct
from tdyno.mnt.mnt_ui import MntUI1S1S


class MntArrPoyn:
    def __init__(self, dt, td=None, omin=None, omax=None, n_o=None, nmf=None, ref_spctrm=None, n=None):
        """
        Poynting flux monitor at an array of points. Records both the real time and the frequency space Poynting vectors.

        Parameters
        ----------
        dt : float
        td : int
        omin : float
        omax : float
        n_o : int
        nmf : float
        ref_spctrm : np.ndarray
        """
        if n is None:
            n = 1
        self.n = n
        self._ss = [SigSpct(dt, td, omin, omax, n_o, nmf, ref_spctrm, n=n) for i in range(6)]  # length is 6, corresponding to 3 E and 3 H components
        self._s = []
        self._S = None

    @property
    def s(self):
        """
        Real time Poynting vectors.

        Returns
        -------
        s : list[tuple[np.ndarray, np.ndarray, np.ndarray]]
        """
        return self._s

    @property
    def S(self):
        """
        Poynting flux in frequency space on the point.

        Returns
        -------
        S : tuple[np.ndarray, np.ndarray, np.ndarray]
            each is a complex 1d array
        """
        return self._S

    @property
    def omg(self):
        """
        frequencies

        Returns
        -------
        omg : np.ndarray, 1d array
        """
        return self._ss[0].omg

    @property
    def t(self):
        """
        List of real time points.

        Returns
        -------
        t : list[float]
        """
        t = None
        for i in range(6):
            if self._ss[i].ts:
                t = self._ss[i].ts
                break
        return t

    def rnf(self, ex=None, ey=None, ez=None, hx=None, hy=None, hz=None):
        for i, f in enumerate([ex, ey, ez, hx, hy, hz]):
            if f is not None:
                self._ss[i].rnf(f)

        _ex, _ey, _ez, _hx, _hy, _hz = [0. if (f is None) else f for f in [ex, ey, ez, hx, hy, hz]]

        sx = (_ey * _hz - _ez * _hy)
        sy = (_ez * _hx - _ex * _hz)
        sz = (_ex * _hy - _ey * _hx)
        self._s.append((sx, sy, sz))

        Ex, Ey, Ez, Hx, Hy, Hz = [self._ss[i].F if (f is not None) else 0. for i, f in enumerate([ex, ey, ez, hx, hy, hz])]
        Exc, Eyc, Ezc, Hxc, Hyc, Hzc = [f.conjugate() for f in [Ex, Ey, Ez, Hx, Hy, Hz]]
        sx = 1./2 * (Ey * Hzc - Ez * Hyc).real
        sy = 1./2 * (Ez * Hxc - Ex * Hzc).real
        sz = 1./2 * (Ex * Hyc - Ey * Hxc).real
        self._S = (sx, sy, sz)

    def reset(self):
        self._s = []
        self._S = None
        for s in self._ss:
            s.reset()


class Mnt2DSqPoyn:
    # todo: take modal profile, get Poynting in specific mode
    # todo: get Poynting flux rate in moving time-window
    # todo: left edge poynting flux is negative!

    def __init__(self,
                 st,
                 xmin=None, xmax=None, ymin=None, ymax=None,
                 dt=None, td=None,
                 omin=None, omax=None, n_o=None,
                 nmf=None, ref_spctrm=None,
                 whr='all', plrz=None):
        """
        Poynting flux through a square in 2D, from inside to outside.

        This monitor records the total real-time flux through the boundary of the box. It also calculates the Poynting flux spectrum.

        You can feed new field data into the monitor through `rnf()`. If polarization is "Ez", the arguments to `rnf()` is understood as Ez, Hx, Hy. For "Hz" polarization, it is understood as Hz, Ex, Ey.

        Parameters
        ----------
        st: S2T
        xmin : float
        xmax : float
        ymin : float
        ymax : float
        dt : float
        td : int
        omin : float
        omax : float
        n_o : int
        nmf : float
        ref_spctrm : np.ndarray
        whr : str
            Controls which sides of the TFSF source exist.
            
            Default to be 'all'. Can be 't', 'b', 'l', or 'r' (top, bottom, left, or right).
        """

        xmin, xmax, ymin, ymax = [smm if mm is None else mm for (mm, smm) in zip([xmin, xmax, ymin, ymax], [st.xmin, st.xmax, st.ymin, st.ymax])]

        # same as TFSF indexing
        xx_n, yy_n = [st.xx_n, st.yy_n]
        self.xx_n, self.yy_n = (xx_n, yy_n)
        xmin_n, xmax_n, ymin_n, ymax_n = [int(np.floor(cor / dcor)) for cor, dcor in zip([xmin + st.dx / 1e4 - st.xmin, xmax + st.dx / 1e4 - st.xmin, ymin + st.dx / 1e4 - st.ymin, ymax + st.dx / 1e4 - st.ymin], [st.dx, st.dx, st.dy, st.dy])]
        if xmin_n < xx_n.min():
            warn('Source left boundary outside of the solving space. It has been reset to the left edge of the solving space.')
            xmin_n = xx_n.min()
        if xmax_n > xx_n.max():
            warn('Source right boundary outside of the solving space. It has been reset to the right edge of the solving space.')
            xmax_n = xx_n.max()
        if ymin_n < yy_n.min():
            warn('Source bottom boundary outside of the solving space. It has been reset to the bottom edge of the solving space.')
            ymin_n = yy_n.min()
        if ymax_n > yy_n.max():
            warn('Source top boundary outside of the solving space. It has been reset to the top edge of the solving space.')
            ymax_n = yy_n.max()
        self.xmin_n, self.xmax_n, self.ymin_n, self.ymax_n = [xmin_n, xmax_n, ymin_n, ymax_n]

        # indexing of the edge in the bulk
        # all 2d arrays
        self.i_te_TM = (yy_n == ymax_n) * (xx_n <= xmax_n) * (xx_n > xmin_n)
        self.i_be_TM = (yy_n == ymin_n) * (xx_n <= xmax_n) * (xx_n > xmin_n)
        self.i_le_TM = (xx_n == xmin_n) * (yy_n <= ymax_n) * (yy_n > ymin_n)
        self.i_re_TM = (xx_n == xmax_n) * (yy_n <= ymax_n) * (yy_n > ymin_n)
        #
        self.i_te_TE = (yy_n == ymax_n) * (xx_n < xmax_n) * (xx_n >= xmin_n)
        self.i_be_TE = (yy_n == ymin_n) * (xx_n < xmax_n) * (xx_n >= xmin_n)
        self.i_le_TE = (xx_n == xmin_n) * (yy_n < ymax_n) * (yy_n >= ymin_n)
        self.i_re_TE = (xx_n == xmax_n) * (yy_n < ymax_n) * (yy_n >= ymin_n)
        #
        self.i_e_TM = self.i_re_TM + self.i_le_TM + self.i_te_TM + self.i_be_TM
        self.i_e_TE = self.i_re_TE + self.i_le_TE + self.i_te_TE + self.i_be_TE

        # monitor edge indexing
        if plrz is None:
            plrz = 'Ez'
        self.plrz = plrz
        self.i = []
        self.n = []  # norm direction
        self.d = []  # dx or dy
        if self.plrz == 'Hz':
            if whr == 'all':
                self.i = [self.i_te_TE, self.i_be_TE, self.i_le_TE, self.i_re_TE]
                self.n = [(0, 1), (0, -1), (-1, 0), (0, 1)]
                self.d = [st.dx, st.dx, st.dy, st.dy]
            else:
                for w in whr:
                    if w == 't':
                        self.i.append(self.i_te_TE)
                        self.n.append((0, 1))
                        self.d.append(st.dx)
                    elif w == 'b':
                        self.i.append(self.i_be_TE)
                        self.n.append((0, -1))
                        self.d.append(st.dx)
                    elif w == 'l':
                        self.i.append(self.i_le_TE)
                        self.n.append((-1, 0))
                        self.d.append(st.dy)
                    elif w == 'r':
                        self.i.append(self.i_re_TE)
                        self.n.append((0, 1))
                        self.d.append(st.dy)
        elif self.plrz == 'Ez':
            if whr == 'all':
                self.i = [self.i_te_TM, self.i_be_TM, self.i_le_TM, self.i_re_TM]
                self.n = [(0, 1), (0, -1), (-1, 0), (0, 1)]
            else:
                for w in whr:
                    if w == 't':
                        self.i.append(self.i_te_TM)
                        self.n.append((0, 1))
                        self.d.append(st.dx)
                    elif w == 'b':
                        self.i.append(self.i_be_TM)
                        self.n.append((0, -1))
                        self.d.append(st.dx)
                    elif w == 'l':
                        self.i.append(self.i_le_TM)
                        self.n.append((-1, 0))
                        self.d.append(st.dy)
                    elif w == 'r':
                        self.i.append(self.i_re_TM)
                        self.n.append((0, 1))
                        self.d.append(st.dy)
        else:
            warn('Polarization not recognized. No monitor value recorded.')
            self.i = [np.zeros([st.Ny, st.Nx], dtype=bool)]  # todo: should here be a list or not?

        self.mpps = [MntArrPoyn(dt, td, omin, omax, n_o, nmf, ref_spctrm, n=np.where(j)[0].size) for j in self.i]  # list, length equal to number of edges included, each element corresponds to one edge
        self.s = []
        self.S = None

        # patch
        self.pchs = []
        if whr == 'all':
            xmi = xmin
            ymi = ymin
            width = xmax - xmin
            height = ymax - ymin
            pch = {'shp': 'rct',
                   'xy': (xmi, ymi),
                   'width': width,
                   'height': height
                   }
            self.pchs.append(pch)
        else:
            for w in whr:
                if w == 'l':
                    xmi = xmin
                    ymi = ymin
                    width = 0.
                    height = ymax - ymin
                elif w == 'b':
                    xmi = xmin
                    ymi = ymin
                    width = xmax - xmin
                    height = 0.
                elif w == 'r':
                    xmi = xmax
                    ymi = ymin
                    width = 0.
                    height = ymax - ymin
                elif w == 't':
                    xmi = xmin
                    ymi = ymax
                    width = xmax - xmin
                    height = 0.
                else:
                    warn('Monitor location unrecognized.', UserWarning)
                    xmi = ymi = width = height = None

                pch = {'shp': 'rct',
                       'xy': (xmi, ymi),
                       'width': width,
                       'height': height
                       }

                self.pchs.append(pch)

    @property
    def omg(self):
        return self.mpps[0].omg

    @property
    def t(self):
        return self.mpps[0].t

    def rnf(self, fz, fx, fy):
        """
        Receive new fields

        Parameters
        ----------
        fx : np.ndarray
        fy : np.ndarray
        fz : np.ndarray
            all of these inputs have shape (Ny, Nx)
        """

        # pick out the field values at the monitor edges
        _fx, _fy, _fz, = [[f[j] for j in self.i] for f in [fx, fy, fz]]  # each is list of 1D arrays, each array corresponds to one edge

        s = 0.
        S = 0.
        for m, x, y, z, n in zip(self.mpps, _fx, _fy, _fz, self.n):
            if self.plrz == "Ez":
                m.rnf(hx=x, hy=y, ez=z)
            elif self.plrz == 'Hz':
                m.rnf(ex=x, ey=y, hz=z)
            else:
                warn("Polarization not recognized. Default to 'Hz'.")
                m.rnf(ex=x, ey=y, hz=z)
            for _s, _S, _n, _d in zip(m.s[-1][:2], m.S[:2], self.n, self.d):
                if _n != 0:
                    s += np.sum(_s) * _d
                    S += np.sum(_S, axis=0) * _d

        self.s.append(s)
        self.S = S

    def cf(self):
        self.s = []
        self.S = None
        for m in self.mpps:
            m.reset()


class Mnt2DSqPoynU:
    def __init__(self, *args, **kwargs):
        """ Wrapper class, `Mnt2DSqPoyn` with `MntUI1S1S`. """
        self.mnt: Mnt2DSqPoyn = Mnt2DSqPoyn(*args, **kwargs)
        self.mui = MntUI1S1S(signal_y_label="Poynting flux (real time)", spectrum_y_label="Poynting flux spectrum", Sx=self.mnt.omg)

    @property
    def pchs(self):
        return self.mnt.pchs

    def rnf(self, *args, **kwargs):
        self.mnt.rnf(*args, **kwargs)

    def up(self):
        self.mui.up(sx=self.mnt.t,
                    sy=self.mnt.s,
                    Sy=self.mnt.S)

    def cf(self):
        self.mnt.cf()



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    readme = f.read()

setup(
    name='tdyno',
    version='0.1.5',
    description='FDTD with dynamic modulations in the refractive index or the gain/loss in materials.',
    long_description=readme,
    long_description_content_type="text/markdown",
    url='https://github.com/alexysong/tdyno',
    # url='http://',
    # check https://pypi.org/pypi?%3Aaction=list_classifiers for classifiers
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
    ],
    keywords=['fdtd', 'dynamic modulation'],
    author='Alex Y. Song',
    author_email='song.alexy@gmail.com',
    packages=find_packages(),
    install_requires=[
        'numpy >= 1.11.3',
        'scipy >= 0.19.0',
        'matplotlib >= 2.0.0'
    ],
    zip_safe=False)



# -*- coding: utf-8 -*-


import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from warnings import warn

from tdyno.s2t import S2T
from tdyno.mnt.sig_spct import SigSpct
from tdyno.mnt.mnt_ui import MntUI1S1S


class MntPntAmp:
    def __init__(self, st, dt=1., td=0, x=0., y=0., omin=0., omax=None, n_o=1000, nmf=1., psd=True, flx=False, ref_spctrm=None):
        """
        monitor the amplitude of field on a point. Record the amplitude, and do Fourier Transform.

        Parameters
        ----------
        st          :   structure
        dt          :   float
                        time resolution
        td          :   int
                        time delay. Won't start Fourier Transform until this time step.
        x, y        :   float
                        the coordinates of the monitor
        omin, omax  :   float
                        omega min and max for the spectrum
        n_o         :   int
                        number of frequency points
        nmf         :   float
                        normalization factor for spectrum
        psd         :   bool
                        if true, plot the power spectral density.
        flx         :   bool
                        if true, plot the flux. Overrides psd.
        ref_spctrm  :   ndarray[float]
                        reference spectrum. If supplied, the shown spectrum will be divided by this.
        """
        warn("This class is deprecated. Use `MntMltPntAmp` instead.", DeprecationWarning)
        self.st = st
        self.dt = dt
        self.td = td
        self.nmf = nmf
        self.psd = psd
        self.flx = flx
        self.ref_spctrm = ref_spctrm

        if x >= st.xmax:
            x = st.xmax - st.xres  #todo if x=st.xmax, still cause weird error
        if x < st.xmin:
            x = st.xmin
        if y >= st.ymax:
            y = st.ymax - st.xres
        if y < st.ymin:
            y = st.ymin
        x_n = int(round((x - st.xmin) / st.xres))
        y_n = int(round((y - st.ymin) / st.yres))
        self.idx_mnt = (st.xx_n == x_n) * (st.yy_n == y_n)

        # current time step
        self.nt = 0

        # list of time steps
        self.t_n = np.array([])

        # Initialize recorded field at the location of the monitor
        self.f_mnt = np.array([])

        # frequency space
        self.omin = omin
        if omax is None:
            self.omax = np.pi / self.dt
        else:
            self.omax = omax
        self.omg = np.linspace(self.omin, self.omax, n_o)

        # calculate the Fourier transform kernel
        self.knl = np.exp(-1j * self.omg * self.dt)

        # Initialize the Fourier transform of the field
        self.F_mnt = np.zeros(n_o) * 1j
        self.F_mnt_sq = np.zeros(n_o)
        self.F_mnt_flx = np.zeros(n_o)
        self.F_shn = np.zeros(n_o)

        # Start plotter
        mpl.rcParams['mathtext.fontset'] = 'cm'
        fontsize_label = 20
        fontsize_axis = 16
        self.fig = plt.figure(figsize=(14, 5))
        ax_f = self.fig.add_subplot(1, 2, 1)
        ax_f.set_xlabel('Time', fontsize=fontsize_label)
        ax_f.set_ylabel('Field Intensity', fontsize=fontsize_label)
        ax_f.tick_params(axis='both', which='major', labelsize=fontsize_axis)
        self.ax_f = ax_f
        self.ln_f = ax_f.plot(self.t_n*self.dt, self.f_mnt)[0]

        ax_F = self.fig.add_subplot(1, 2, 2)
        ax_F.set_xlabel('$\omega$ ($2\pi$)', fontsize=fontsize_label)
        if self.flx:
            ax_F.set_ylabel('Flux', fontsize=fontsize_label)
        else:
            ax_F.set_ylabel('Power Spectral Density', fontsize=fontsize_label)
        ax_F.tick_params(axis='both', which='major', labelsize=fontsize_axis)
        self.ax_F = ax_F
        self.ln_F = ax_F.plot(self.omg/2./np.pi, np.square(np.abs(self.F_mnt)))[0]
        ax_F.set_xlim([self.omg.min()/2./np.pi, self.omg.max()/2./np.pi])

        self.pch = {'shp': 'rct',
                    'xy': (x, y),
                    'width': st.dx,
                    'height': st.dy
                    }
        self.pchs = [self.pch]

    def rnf(self, fz, fx, fy):
        """
        receive new field

        Parameters
        ----------
        fz          :   ndarray

        Returns
        -------

        """
        self.t_n = np.append(self.t_n, self.nt)

        f_mnt = fz[self.idx_mnt]
        self.f_mnt = np.append(self.f_mnt, f_mnt)

        if self.nt > self.td:
            self.F_mnt += (self.knl ** self.nt) * f_mnt * self.dt
            self.F_mnt_sq = np.square(np.abs(self.F_mnt)) / self.nmf
            if self.flx:
                self.F_mnt_flx = self.F_mnt_sq / self.omg
                self.F_shn = self.F_mnt_flx
            else:
                self.F_shn = self.F_mnt_sq
            if self.ref_spctrm is not None:
                self.F_shn /= self.ref_spctrm

        self.nt += 1

    def up(self):
        """
        update plots

        Returns
        -------

        """
        self.ln_f.set_data(self.t_n*self.dt, self.f_mnt)
        xmin = self.t_n.min() * self.dt
        xmax = self.t_n.max() * self.dt
        ymin = self.f_mnt.min()
        ymax = self.f_mnt.max()
        self.ax_f.set_xlim([xmin, xmax])
        self.ax_f.set_ylim([ymin, ymax])

        self.ax_f.draw_artist(self.ax_f.patch)
        self.ax_f.draw_artist(self.ln_f)
        self.fig.canvas.flush_events()

        self.ln_F.set_ydata(self.F_shn)
        self.ax_F.set_ylim([self.F_shn.min(), self.F_shn.max()])

        self.ax_F.draw_artist(self.ax_F.patch)
        self.ax_F.draw_artist(self.ln_F)
        self.fig.canvas.flush_events()

        self.fig.canvas.update()

    def cf(self):
        """
        clear fields.

        Returns
        -------

        """
        self.nt = 0
        self.t_n = np.array([])
        self.f_mnt = np.array([])
        self.F_mnt = np.zeros(self.F_mnt.size) * 1j


class MntMltPntAmp:
    def __init__(self, st, coords, dt=None, td=None, wts=None, omin=None, omax=None, n_o=None, nmf=None, show=None, ref_spctrm=None):
        """
        Monitor the amplitude of field on multi points. Record the amplitudes on each, do weighted sum, and do Fourier Transform.

        Parameters
        ----------
        st          :   S2T
        dt          :   float
                        time resolution
        td          :   int
                        time delay. Won't start Fourier Transform until this time step.
        coords      :   list[tuple[float, float]]
                        coordinates of the points to monitor. Each element is a tuple of (x, y)
        wts         :   list
                        weights, the amplitudes of the monitored points are multiplied by these weights, and then add up
        omin, omax  :   float
                        omega min and max for the spectrum
        n_o         :   int
                        number of frequency points
        nmf         :   float
                        normalization factor for spectrum
        show        :   str
                        {'energy spectral density', 'power spectral density', 'flux spectral density', 'flux rate spectral density'}

                        showing what kind of Fourier Transform of the field.

                        "flux" means photon flux, i.e. dividing intensity by frequency.

        if_esd      :   bool
                        if true, plot the energy spectral density.
        if_psd      :   bool
                        if true, plot the energy spectral density.
        if_flx      :   bool
                        if true, plot the flux. Overrides if_esd.
        ref_spctrm  :   ndarray[float]
                        reference spectrum. If supplied, the shown spectrum will be divided by this.
        """
        if wts is None:
            wts = [1.]
        self.wts = wts

        if (show != 'energy spectral density') and (show != 'power spectral density') and (show != 'flux spectral density') and (show != 'flux rate spectral density'):
            warn('the choice of the spectrum to show not understood. Default to energy spectral density', UserWarning)
            show = 'energy spectral density'
        self.show = show

        self.coords = []
        self.idx_mnts = []
        self.f_mnts = []
        if coords is None:
            coords = [(0., 0.)]
        if not hasattr(coords[0], '__len__'):
            coords = [coords]
        for (x, y) in coords:
            if x >= st.xmax:
                x = st.xmax - st.xres
            if x < st.xmin:
                x = st.xmin
            if y >= st.ymax:
                y = st.ymax - st.xres
            if y < st.ymin:
                y = st.ymin
            self.coords.append((x, y))

            x_n = int(round((x - st.xmin) / st.xres))
            y_n = int(round((y - st.ymin) / st.yres))
            idx_mnt = (st.xx_n == x_n) * (st.yy_n == y_n)
            self.idx_mnts.append(idx_mnt)

        # Initialize weighted sum of recorded field
        self.s_ws = SigSpct(dt, td, omin, omax, n_o, nmf, ref_spctrm)

        self.F = None  # spectrum shown

        self.mui = MntUI1S1S(signal_y_label="Field strength", spectrum_y_label=self.show, Sx=self.s_ws.omg)  # monitor UI

        self.pchs = []
        for (x, y) in self.coords:
            pch = {'shp': 'rct',
                    'xy': (x, y),
                    'width': st.dx,
                    'height': st.dy
                    }
            self.pchs.append(pch)

    def rnf(self, fz, fx, fy):
        """
        Receive new field

        Parameters
        ----------
        f           :   ndarray

        Returns
        -------

        """
        f_n_ws = 0.
        for idx, w in zip(self.idx_mnts, self.wts):
            f_n_ws += w * fz[idx]
        self.s_ws.rnf(f_n_ws)

    def up(self):
        """
        Update plots

        Returns
        -------

        """

        if self.show == 'power spectral density':
            F = self.s_ws.psd
        elif self.show == 'flux spectral density':
            F = self.s_ws.flx
        elif self.show == 'flux rate spectral density':
            F = self.s_ws.flxr
        else:
            F = self.s_ws.esd
        self.F = F

        self.mui.up(sx=self.s_ws.ts, sy=self.s_ws.f, Sy=self.F.ravel())

    def cf(self):
        """
        Clear fields.
        """
        self.s_ws.reset()
